"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.defaultConfig = exports.defaultBrowser = void 0;
exports.readConfig = readConfig;
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const url_1 = require("url");
const browser_js_1 = require("./providers/browser.js");
const types_js_1 = require("../types.js");
const utils_js_1 = require("./utils.js");
const reporter_js_1 = require("./reporter.js");
const logger_js_1 = require("./logger.js");
exports.defaultBrowser = 'chrome';
exports.defaultConfig = {
    disableTelemetry: false,
    useDocker: true,
    dockerImage: 'aerokube/selenoid:latest-release',
    dockerImagePlatform: '',
    pullImages: true,
    failFast: false,
    storybookUrl: 'http://localhost:6006',
    screenDir: path_1.default.resolve('images'),
    reportDir: path_1.default.resolve('report'),
    reporter: process.env.TEAMCITY_VERSION ? reporter_js_1.TeamcityReporter : reporter_js_1.CreeveyReporter,
    storiesProvider: browser_js_1.loadStories,
    maxRetries: 0,
    testTimeout: 30000,
    diffOptions: { threshold: 0.05, includeAA: false },
    odiffOptions: { threshold: 0.05, antialiasing: true },
    browsers: { [exports.defaultBrowser]: true },
    hooks: {},
    testsRegex: /\.creevey\.(t|j)s$/,
};
function normalizeBrowserConfig(name, config) {
    if (typeof config == 'boolean')
        return { browserName: name };
    if (typeof config == 'string')
        return { browserName: config };
    return config;
}
function resolveConfigPath(configPath) {
    const configDir = path_1.default.resolve('.creevey');
    if ((0, types_js_1.isDefined)(configPath)) {
        configPath = path_1.default.resolve(configPath);
    }
    else if (fs_1.default.existsSync(configDir)) {
        for (const ext of utils_js_1.configExt) {
            configPath = path_1.default.resolve(configDir, `config${ext}`);
            if (fs_1.default.existsSync(configPath))
                break;
        }
    }
    else {
        for (const ext of utils_js_1.configExt) {
            configPath = path_1.default.resolve(`creevey.config${ext}`);
            if (fs_1.default.existsSync(configPath))
                break;
        }
    }
    return configPath;
}
async function readConfig(options) {
    const configPath = resolveConfigPath(options.config);
    const userConfig = { ...exports.defaultConfig };
    if ((0, types_js_1.isDefined)(configPath)) {
        const configModule = await (0, utils_js_1.loadThroughTSX)((load) => {
            const configFileUrl = (0, url_1.pathToFileURL)(configPath).toString();
            return load(configFileUrl);
        });
        let configData = 'default' in configModule ? configModule.default : configModule;
        // NOTE In node > 18 with commonjs project and esm config with tsconfig moduleResolution nodeNext there is additional 'default'
        configData = 'default' in configData ? configData.default : configData;
        if (!configData.webdriver) {
            const { SeleniumWebdriver } = await import('./selenium/webdriver.js');
            logger_js_1.logger.warn("Creevey supports `Selenium` and `Playwright` webdrivers. For backward compatibility `Selenium` is used by default, but it might changed in the future. Please explicitly specify one of webdrivers in your Creevey's config");
            configData.webdriver = SeleniumWebdriver;
        }
        Object.assign(userConfig, configData);
    }
    if (options.failFast != undefined)
        userConfig.failFast = Boolean(options.failFast);
    if (options.reportDir)
        userConfig.reportDir = path_1.default.resolve(options.reportDir);
    if (options.screenDir)
        userConfig.screenDir = path_1.default.resolve(options.screenDir);
    if (options.storybookUrl)
        userConfig.storybookUrl = options.storybookUrl;
    // NOTE: Hack to pass typescript checking
    const config = userConfig;
    Object.entries(config.browsers).forEach(([browser, browserConfig]) => (config.browsers[browser] = normalizeBrowserConfig(browser, browserConfig)));
    return config;
}
//# sourceMappingURL=config.js.map