"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.pullImages = pullImages;
exports.buildImage = buildImage;
exports.runImage = runImage;
const tar_stream_1 = __importDefault(require("tar-stream"));
const stream_1 = require("stream");
const dockerode_1 = __importDefault(require("dockerode"));
const messages_js_1 = require("./messages.js");
const logger_js_1 = require("./logger.js");
const docker = new dockerode_1.default();
class DevNull extends stream_1.Writable {
    _write(_chunk, _encoding, callback) {
        setImmediate(callback);
    }
}
async function pullImages(images, { auth, platform } = {}) {
    const args = {};
    if (auth)
        args.authconfig = auth;
    if (platform)
        args.platform = platform;
    logger_js_1.logger.info('Pull docker images');
    // TODO Replace with `import from`
    const { default: yoctoSpinner } = await import('yocto-spinner');
    for (const image of images) {
        await new Promise((resolve, reject) => {
            const spinner = yoctoSpinner({ text: `${image}: Pull start` }).start();
            docker.pull(image, args, (pullError, stream) => {
                if (pullError || !stream) {
                    spinner.error(pullError?.message);
                    reject(pullError ?? new Error('Unknown error'));
                    return;
                }
                docker.modem.followProgress(stream, onFinished, onProgress);
                function onFinished(error) {
                    if (error) {
                        spinner.error(error.message);
                        reject(error);
                        return;
                    }
                    spinner.success(`${image}: Pull complete`);
                    resolve();
                }
                function onProgress(event) {
                    if (!/^[a-z0-9]{12}$/i.test(event.id))
                        return;
                    spinner.text = `${image}: [${event.id}] ${event.status} ${event.progress ? event.progress : ''}`;
                }
            });
        });
    }
}
async function buildImage(imageName, dockerfile) {
    const images = await docker.listImages({ filters: { label: [`creevey=${imageName}`] } });
    if (images.at(0)) {
        logger_js_1.logger.info(`Image ${imageName} already exists`);
        return;
    }
    const pack = tar_stream_1.default.pack();
    pack.entry({ name: 'Dockerfile' }, dockerfile);
    pack.finalize();
    const { default: yoctoSpinner } = await import('yocto-spinner');
    const spinner = yoctoSpinner({ text: `${imageName}: Build start` }).start();
    await new Promise((resolve, reject) => {
        void docker.buildImage(
        // @ts-expect-error Type incompatibility AsyncIterator and AsyncIterableIterator
        pack, { t: imageName, labels: { creevey: imageName } }, (buildError, stream) => {
            if (buildError || !stream) {
                spinner.error(buildError?.message);
                reject(buildError ?? new Error('Unknown error'));
                return;
            }
            docker.modem.followProgress(stream, onFinished, onProgress);
            function onFinished(error) {
                if (error) {
                    spinner.error(error.message);
                    reject(error);
                    return;
                }
                spinner.success(`${imageName}: Build complete`);
                resolve();
            }
            function onProgress(event) {
                if (!/^[a-z0-9]{12}$/i.test(event.id))
                    return;
                spinner.text = `${imageName}: [${event.id}] ${event.status} ${event.progress ? event.progress : ''}`;
            }
        });
    });
}
async function runImage(image, args, options, debug) {
    await Promise.all((await docker.listContainers({ all: true, filters: { ancestor: [image] } })).map(async (info) => {
        const container = docker.getContainer(info.Id);
        try {
            await container.stop();
        }
        catch {
            /* noop */
        }
        await container.remove();
    }));
    const hub = docker.run(image, args, debug ? process.stdout : new DevNull(), options, (error) => {
        if (error)
            throw error;
    });
    return new Promise((resolve) => {
        hub.once('container', (container) => {
            // eslint-disable-next-line @typescript-eslint/no-misused-promises
            (0, messages_js_1.subscribeOn)('shutdown', async () => {
                try {
                    await container.stop();
                    await container.remove();
                }
                catch {
                    /* noop */
                }
            });
        });
        hub.once('start', (container) => void container.inspect().then((info) => {
            resolve(info.NetworkSettings.Networks.bridge.IPAddress);
        }));
    });
}
//# sourceMappingURL=docker.js.map