"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = default_1;
const cluster_1 = __importDefault(require("cluster"));
const config_js_1 = require("./config.js");
const types_js_1 = require("../types.js");
const logger_js_1 = require("./logger.js");
const webdriver_js_1 = require("./selenium/webdriver.js");
const webdriver_js_2 = require("./webdriver.js");
const utils_js_1 = require("./utils.js");
const messages_js_1 = require("./messages.js");
const docker_file_js_1 = require("./playwright/docker-file.js");
const docker_js_1 = require("./docker.js");
async function startWebdriverServer(browser, config, options) {
    if (config.webdriver === webdriver_js_1.SeleniumWebdriver) {
        if (cluster_1.default.isPrimary) {
            const { startSelenoidContainer, startSelenoidStandalone } = await import('./selenium/selenoid.js');
            const gridUrl = 'http://localhost:4444/wd/hub';
            if (config.useDocker) {
                const host = await startSelenoidContainer(config, options.debug);
                return utils_js_1.isInsideDocker ? gridUrl.replace(webdriver_js_2.LOCALHOST_REGEXP, host) : gridUrl;
            }
            await startSelenoidStandalone(config, options.debug);
            return gridUrl;
        }
        // TODO Worker might want to use docker image of browser or start standalone selenium
    }
    else {
        if (config.gridUrl)
            return undefined;
        // TODO start standalone playwright server (useDocker == false)
        const { default: { version }, } = await import('playwright-core/package.json', { with: { type: 'json' } });
        if (cluster_1.default.isWorker) {
            // TODO Re-use dockerImage
            // TODO Use https://hub.docker.com/r/playwright/chrome
            // NOTE It will be possible to use `chrome` browserName
            const { startPlaywrightContainer } = await import('./playwright/docker.js');
            const { browserName } = config.browsers[browser];
            const imageName = `creevey/${browserName}:v${version}`;
            const host = await startPlaywrightContainer(imageName, options.debug);
            return host;
        }
        else {
            const browsers = [...new Set(Object.values(config.browsers).map((c) => c.browserName))];
            await Promise.all(browsers.map(async (browserName) => {
                const imageName = `creevey/${browserName}:v${version}`;
                const dockerfile = (0, docker_file_js_1.playwrightDockerFile)(browserName, version);
                await (0, docker_js_1.buildImage)(imageName, dockerfile);
            }));
            const { default: getPort } = await import('get-port');
            cluster_1.default.on('message', (worker, message) => {
                if (!(0, types_js_1.isWorkerMessage)(message))
                    return;
                const workerMessage = message;
                if (workerMessage.type != 'port')
                    return;
                void getPort().then((port) => {
                    (0, messages_js_1.sendWorkerMessage)(worker, {
                        type: 'port',
                        payload: { port },
                    });
                });
            });
        }
        // TODO Support gridUrl for playwright
        // NOTE: There is no grid for playwright right now
    }
}
async function default_1(options) {
    const config = await (0, config_js_1.readConfig)(options);
    const { browser = config_js_1.defaultBrowser, update, ui, port } = options;
    let gridUrl = cluster_1.default.isPrimary ? config.gridUrl : options.gridUrl;
    // NOTE: We don't need docker nor selenoid for update option
    if (!(gridUrl || Object.values(config.browsers).every(({ gridUrl }) => gridUrl)) &&
        !update) {
        gridUrl = await startWebdriverServer(browser, config, options);
    }
    switch (true) {
        case Boolean(update): {
            (await import('./update.js')).update(config, typeof update == 'string' ? update : undefined);
            return;
        }
        case cluster_1.default.isPrimary: {
            if (config.webdriver === webdriver_js_1.SeleniumWebdriver) {
                try {
                    await import('selenium-webdriver');
                }
                catch {
                    logger_js_1.logger.error('Failed to start Creevey, missing required dependency: "selenium-webdriver"');
                    process.exit(-1);
                }
            }
            else {
                try {
                    await import('playwright-core');
                }
                catch {
                    logger_js_1.logger.error('Failed to start Creevey, missing required dependency: "playwright-core"');
                    process.exit(-1);
                }
            }
            logger_js_1.logger.info('Starting Master Process');
            const resolveApi = (await import('./master/server.js')).start(config.reportDir, port, ui);
            return (await import('./master/start.js')).start(gridUrl, config, options, resolveApi);
        }
        default: {
            logger_js_1.logger.info(`Starting Worker for ${browser}`);
            // NOTE: We assume that we pass `gridUrl` to worker CLI options
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            return (await import('./worker/start.js')).start(browser, gridUrl, config, options);
        }
    }
}
//# sourceMappingURL=index.js.map