"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.InternalBrowser = void 0;
const playwright_core_1 = require("playwright-core");
const loglevel_1 = __importDefault(require("loglevel"));
const chalk_1 = __importDefault(require("chalk"));
const uuid_1 = require("uuid");
const loglevel_plugin_prefix_1 = __importDefault(require("loglevel-plugin-prefix"));
const types_1 = require("../../types");
const messages_1 = require("../messages");
const webdriver_1 = require("../webdriver");
const utils_1 = require("../utils");
const logger_1 = require("../logger");
async function tryConnect(type, gridUrl) {
    let timeout = null;
    let isTimeout = false;
    let error = null;
    return Promise.race([
        new Promise((resolve) => (timeout = setTimeout(() => {
            isTimeout = true;
            logger_1.logger.error(`Can't connect to ${type.name()} playwright browser`, error);
            resolve(null);
        }, 10000))),
        (async () => {
            let browser = null;
            do {
                try {
                    browser = await type.connect(gridUrl);
                    // eslint-disable-next-line @typescript-eslint/no-unnecessary-condition
                    if (timeout)
                        clearTimeout(timeout);
                    break;
                }
                catch (e) {
                    error = e;
                }
                // eslint-disable-next-line @typescript-eslint/no-unnecessary-condition
            } while (!isTimeout);
            return browser;
        })(),
    ]);
}
class InternalBrowser {
    #isShuttingDown = false;
    #browser;
    #page;
    #sessionId = (0, uuid_1.v4)();
    #serverHost = null;
    #serverPort;
    #logger;
    #unsubscribe = types_1.noop;
    constructor(browser, page, port) {
        this.#browser = browser;
        this.#page = page;
        this.#serverPort = port;
        this.#logger = loglevel_1.default.getLogger(this.#sessionId);
        this.#unsubscribe = (0, messages_1.subscribeOn)('shutdown', () => {
            void this.closeBrowser();
        });
    }
    get browser() {
        return this.#page;
    }
    get sessionId() {
        return this.#sessionId;
    }
    async closeBrowser() {
        if (this.#isShuttingDown)
            return;
        this.#isShuttingDown = true;
        this.#unsubscribe();
        try {
            await this.#page.close();
            await this.#browser.close();
        }
        catch (_) {
            /* noop */
        }
    }
    async takeScreenshot(captureElement, ignoreElements) {
        // TODO Implement features from selenium `takeScreenshot`
        // TODO Do we need scroll bar hack from selenium?
        const ignore = Array.isArray(ignoreElements) ? ignoreElements : ignoreElements ? [ignoreElements] : [];
        const mask = ignore.map((el) => this.#page.locator(el));
        if (captureElement) {
            const element = await this.#page.$(captureElement);
            if (!element)
                throw new Error(`Element with selector ${captureElement} not found`);
            return element.screenshot({ animations: 'disabled', mask });
        }
        return this.#page.screenshot({ animations: 'disabled', mask, fullPage: true });
    }
    waitForComplete(callback) {
        void this.#page.evaluate(() => window.__CREEVEY_HAS_PLAY_COMPLETED_YET__()).then(callback);
    }
    async selectStory(id, waitForReady = false) {
        // NOTE: Global variables might be reset after hot reload. I think it's workaround, maybe we need better solution
        await this.updateBrowserGlobalVariables();
        await this.resetMousePosition();
        this.#logger.debug(`Triggering 'SetCurrentStory' event with storyId ${chalk_1.default.magenta(id)}`);
        const result = await this.#page.evaluate(([id, shouldWaitForReady]) => {
            if (typeof window.__CREEVEY_SELECT_STORY__ == 'undefined') {
                return [
                    "Creevey can't switch story. This may happened if forget to add `creevey` addon to your storybook config, or storybook not loaded in browser due syntax error.",
                ];
            }
            return window.__CREEVEY_SELECT_STORY__(id, shouldWaitForReady);
        }, [id, waitForReady]);
        const [errorMessage, isCaptureCalled = false] = result ?? [];
        if (errorMessage)
            throw new Error(errorMessage);
        return isCaptureCalled;
    }
    async updateStoryArgs(story, updatedArgs) {
        await this.#page.evaluate(([storyId, updatedArgs, UPDATE_STORY_ARGS, STORY_RENDERED]) => {
            return new Promise((resolve) => {
                window.__STORYBOOK_ADDONS_CHANNEL__.once(STORY_RENDERED, resolve);
                window.__STORYBOOK_ADDONS_CHANNEL__.emit(UPDATE_STORY_ARGS, {
                    storyId,
                    updatedArgs,
                });
            });
        }, [story.id, updatedArgs, types_1.StorybookEvents.UPDATE_STORY_ARGS, types_1.StorybookEvents.STORY_RENDERED]);
    }
    async loadStoriesFromBrowser(retry = false) {
        try {
            const stories = await this.#page.evaluate(() => window.__CREEVEY_GET_STORIES__());
            if (!stories)
                throw new Error("Can't get stories, it seems creevey or storybook API isn't available");
            return stories;
        }
        catch (error) {
            // TODO Check how other solutions with playwright get stories from storybook
            if (retry)
                throw error;
            await new Promise((resolve) => setTimeout(resolve, 1000));
            // NOTE: Try one more time because of dynamic nature of vite and storybook
            return this.loadStoriesFromBrowser(true);
        }
    }
    static async getBrowser(browserName, gridUrl, config, options) {
        const browserConfig = config.browsers[browserName];
        const { storybookUrl: address = config.storybookUrl, viewport, _storybookGlobals, ...userCapabilities } = browserConfig;
        let browser = null;
        if (new URL(gridUrl).protocol === 'ws:') {
            switch (browserConfig.browserName) {
                case 'chromium':
                    browser = await tryConnect(playwright_core_1.chromium, gridUrl);
                    break;
                case 'firefox':
                    browser = await tryConnect(playwright_core_1.firefox, gridUrl);
                    break;
                case 'webkit':
                    browser = await tryConnect(playwright_core_1.webkit, gridUrl);
                    break;
                default:
                    logger_1.logger.error(`Unknown browser ${browserConfig.browserName}. Playwright supports browsers: chromium, firefox, webkit`);
            }
        }
        else {
            if (browserConfig.browserName != 'chrome') {
                logger_1.logger.error("Playwright's Selenium Grid feature supports only chrome browser");
                return null;
            }
            process.env.SELENIUM_REMOTE_URL = gridUrl;
            process.env.SELENIUM_REMOTE_CAPABILITIES = JSON.stringify(userCapabilities);
            browser = await playwright_core_1.chromium.launch();
        }
        if (!browser) {
            return null;
        }
        const page = await browser.newPage();
        // TODO Add debug output
        const internalBrowser = new InternalBrowser(browser, page, options.port);
        try {
            if (utils_1.isShuttingDown.current)
                return null;
            const done = await internalBrowser.init({
                browserName,
                viewport,
                storybookUrl: address,
                storybookGlobals: _storybookGlobals,
                resolveStorybookUrl: config.resolveStorybookUrl,
            });
            return done ? internalBrowser : null;
        }
        catch (originalError) {
            void internalBrowser.closeBrowser();
            const message = originalError instanceof Error ? originalError.message : originalError;
            const error = new Error(`Can't load storybook root page: ${message}`);
            if (originalError instanceof Error)
                error.stack = originalError.stack;
            logger_1.logger.error(error);
            return null;
        }
    }
    async init({ browserName, viewport, storybookUrl, storybookGlobals, resolveStorybookUrl, }) {
        const sessionId = this.#sessionId;
        loglevel_plugin_prefix_1.default.apply(this.#logger, {
            format(level) {
                const levelColor = logger_1.colors[level.toUpperCase()];
                return `[${browserName}:${chalk_1.default.gray(sessionId)}] ${levelColor(level)} =>`;
            },
        });
        this.#page.setDefaultNavigationTimeout(10000);
        this.#page.setDefaultTimeout(60000);
        return await (0, utils_1.runSequence)([
            () => this.openStorybookPage(storybookUrl, resolveStorybookUrl),
            () => this.waitForStorybook(),
            () => this.updateStorybookGlobals(storybookGlobals),
            () => this.resolveCreeveyHost(),
            () => this.updateBrowserGlobalVariables(),
            () => this.resizeViewport(viewport),
        ], () => !this.#isShuttingDown);
    }
    async openStorybookPage(storybookUrl, resolver) {
        if (!webdriver_1.LOCALHOST_REGEXP.test(storybookUrl)) {
            await this.#page.goto((0, webdriver_1.appendIframePath)(storybookUrl));
            return;
        }
        try {
            if (resolver) {
                this.#logger.debug('Resolving storybook url with custom resolver');
                const resolvedUrl = await resolver();
                this.#logger.debug(`Resolver storybook url ${resolvedUrl}`);
                await this.#page.goto((0, webdriver_1.appendIframePath)(resolvedUrl));
            }
            else {
                await (0, webdriver_1.resolveStorybookUrl)((0, webdriver_1.appendIframePath)(storybookUrl), (url) => this.checkUrl(url), this.#logger);
            }
        }
        catch (error) {
            this.#logger.error('Failed to resolve storybook URL', error instanceof Error ? error.message : '');
            throw error;
        }
    }
    async checkUrl(url) {
        try {
            this.#logger.debug(`Opening ${chalk_1.default.magenta(url)} and checking the page source`);
            const response = await this.#page.goto(url, { waitUntil: 'commit' });
            const source = await response?.text();
            this.#logger.debug(`Checking ${chalk_1.default.cyan(`#${webdriver_1.storybookRootID}`)} existence on ${chalk_1.default.magenta(url)}`);
            return source?.includes(`id="${webdriver_1.storybookRootID}"`) ?? false;
        }
        catch {
            return false;
        }
    }
    async waitForStorybook() {
        // TODO Duplicated code with selenium
        this.#logger.debug('Waiting for `setStories` event to make sure that storybook is initiated');
        const isTimeout = await Promise.race([
            new Promise((resolve) => {
                setTimeout(() => {
                    resolve(true);
                }, 60000);
            }),
            (async () => {
                let wait = true;
                do {
                    try {
                        // TODO Research a different way to ensure storybook is initiated
                        wait = await this.#page.evaluate((SET_GLOBALS) => {
                            if (typeof window.__STORYBOOK_ADDONS_CHANNEL__ == 'undefined')
                                return true;
                            if (window.__STORYBOOK_ADDONS_CHANNEL__.last(SET_GLOBALS) == undefined)
                                return true;
                            return false;
                        }, types_1.StorybookEvents.SET_GLOBALS);
                    }
                    catch (e) {
                        this.#logger.debug('An error has been caught during the script:', e);
                    }
                } while (wait);
                return false;
            })(),
        ]);
        // TODO Change the message to describe a reason why it might happen
        if (isTimeout)
            throw new Error('Failed to wait `setStories` event');
    }
    async updateStorybookGlobals(globals) {
        if (!globals)
            return;
        this.#logger.debug('Applying storybook globals');
        await this.#page.evaluate((globals) => {
            window.__CREEVEY_UPDATE_GLOBALS__(globals);
        }, globals);
    }
    async resolveCreeveyHost() {
        const addresses = (0, webdriver_1.getAddresses)();
        this.#serverHost = await this.#page.evaluate(([hosts, port]) => {
            return Promise.all(hosts.map((host) => {
                return Promise.race([
                    // eslint-disable-next-line @typescript-eslint/restrict-plus-operands
                    fetch('http://' + host + ':' + port + '/ping').then((response) => response.text()),
                    new Promise((_resolve, reject) => {
                        setTimeout(reject, 5000);
                    }),
                ])
                    .then((pong) => (pong == 'pong' ? host : null))
                    .catch(() => null);
            })).then((hosts) => hosts.find((host) => host != null) ?? null);
        }, [addresses, this.#serverPort]);
        if (this.#serverHost == null)
            throw new Error("Can't reach creevey server from a browser");
    }
    async updateBrowserGlobalVariables() {
        await this.#page.evaluate(([workerId, creeveyHost, creeveyPort]) => {
            window.__CREEVEY_WORKER_ID__ = workerId;
            window.__CREEVEY_SERVER_HOST__ = creeveyHost ?? 'localhost';
            window.__CREEVEY_SERVER_PORT__ = creeveyPort;
        }, [process.pid, this.#serverHost, this.#serverPort]);
    }
    async resizeViewport(viewport) {
        if (!viewport)
            return;
        await this.#page.setViewportSize(viewport);
    }
    async resetMousePosition() {
        await this.#page.mouse.move(0, 0);
    }
}
exports.InternalBrowser = InternalBrowser;
//# sourceMappingURL=internal.js.map