"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.InternalBrowser = void 0;
const chalk_1 = __importDefault(require("chalk"));
const http_1 = __importDefault(require("http"));
const https_1 = __importDefault(require("https"));
const loglevel_1 = __importDefault(require("loglevel"));
const loglevel_plugin_prefix_1 = __importDefault(require("loglevel-plugin-prefix"));
const pngjs_1 = require("pngjs");
const selenium_webdriver_1 = require("selenium-webdriver");
// import { Options as IeOptions } from 'selenium-webdriver/ie';
// import { Options as EdgeOptions } from 'selenium-webdriver/edge';
// import { Options as ChromeOptions } from 'selenium-webdriver/chrome';
// import { Options as SafariOptions } from 'selenium-webdriver/safari';
// import { Options as FirefoxOptions } from 'selenium-webdriver/firefox';
const capabilities_js_1 = require("selenium-webdriver/lib/capabilities.js");
const types_js_1 = require("../../types.js");
const logger_js_1 = require("../logger.js");
const messages_js_1 = require("../messages.js");
const utils_js_1 = require("../utils.js");
const webdriver_js_1 = require("../webdriver.js");
// type UnPromise<P> = P extends Promise<infer T> ? T : never;
// let context: UnPromise<ReturnType<typeof BrowsingContext>> | null = null;
function getSessionData(grid, sessionId = '') {
    const gridUrl = new URL(grid);
    gridUrl.pathname = `/host/${sessionId}`;
    return new Promise((resolve, reject) => (gridUrl.protocol == 'https:' ? https_1.default : http_1.default).get(gridUrl.toString(), (res) => {
        if (res.statusCode !== 200) {
            reject(new Error(`Couldn't get session data for ${sessionId}. Status code: ${res.statusCode ?? 'Unknown'}`));
            return;
        }
        let data = '';
        res.setEncoding('utf-8');
        res.on('data', (chunk) => (data += chunk));
        res.on('end', () => {
            try {
                resolve(JSON.parse(data));
            }
            catch (error) {
                reject(new Error(`Couldn't get session data for ${sessionId}. ${error instanceof Error ? (error.stack ?? error.message) : error}`));
            }
        });
    }));
}
async function openUrlAndWaitForPageSource(browser, url, predicate) {
    let source = '';
    await browser.get(url);
    do {
        try {
            source = await browser.getPageSource();
        }
        catch {
            // NOTE: Firefox can raise exception "curContainer.frame.document.documentElement is null"
        }
    } while (predicate(source));
    return source;
}
async function buildWebdriver(browserName, gridUrl, config, options) {
    const browserConfig = config.browsers[browserName];
    const { _storybookGlobals, /*customizeBuilder,*/ ...userCapabilities } = browserConfig;
    const url = new URL(gridUrl);
    url.username = url.username ? '********' : '';
    url.password = url.password ? '********' : '';
    logger_js_1.logger.debug(`(${browserName}) Connecting to Selenium ${chalk_1.default.magenta(url.toString())}`);
    // TODO Define some capabilities explicitly and define typings
    const capabilities = new selenium_webdriver_1.Capabilities({ ...userCapabilities, pageLoadStrategy: capabilities_js_1.PageLoadStrategy.EAGER });
    const prefs = new selenium_webdriver_1.logging.Preferences();
    if (options.trace) {
        for (const type of Object.values(selenium_webdriver_1.logging.Type)) {
            prefs.setLevel(type, selenium_webdriver_1.logging.Level.ALL);
        }
    }
    // TODO Fetch selenium grid capabilities
    // TODO Validate browsers, versions, and platform
    // TODO Use `customizeBuilder`
    let browser;
    try {
        // const ie = new IeOptions();
        // const edge = new EdgeOptions();
        // const chrome = new ChromeOptions();
        // const safari = new SafariOptions();
        // const firefox = new FirefoxOptions();
        // edge.enableBidi();
        // chrome.enableBidi();
        // firefox.enableBidi();
        browser = await new selenium_webdriver_1.Builder()
            // .setIeOptions(ie)
            // .setEdgeOptions(edge)
            // .setChromeOptions(chrome)
            // .setSafariOptions(safari)
            // .setFirefoxOptions(firefox)
            .usingServer(gridUrl)
            .withCapabilities(capabilities)
            .setLoggingPrefs(prefs) // NOTE: Should go last
            .build();
        // const id = await browser.getWindowHandle();
        // context = await BrowsingContext(browser, { browsingContextId: id });
    }
    catch (error) {
        logger_js_1.logger.error(`(${browserName}) Failed to start browser:`, error);
        return null;
    }
    return browser;
}
class InternalBrowser {
    #isShuttingDown = false;
    #browser;
    #serverHost = null;
    #serverPort;
    #logger;
    #unsubscribe = types_js_1.noop;
    #keepAliveInterval = null;
    constructor(browser, port, logger) {
        this.#browser = browser;
        this.#serverPort = port;
        this.#logger = logger;
        this.#unsubscribe = (0, messages_js_1.subscribeOn)('shutdown', () => {
            void this.closeBrowser();
        });
    }
    get browser() {
        return this.#browser;
    }
    async closeBrowser() {
        if (this.#isShuttingDown)
            return;
        this.#isShuttingDown = true;
        this.#unsubscribe();
        if (this.#keepAliveInterval !== null)
            clearInterval(this.#keepAliveInterval);
        try {
            await this.#browser.quit();
        }
        catch (_) {
            /* noop */
        }
    }
    async takeScreenshot(captureElement, ignoreElements) {
        let screenshot;
        const ignoreStyles = await this.insertIgnoreStyles(ignoreElements);
        if (this.#logger.getLevel() <= loglevel_1.default.levels.DEBUG) {
            const { innerWidth, innerHeight } = await this.#browser.executeScript(function () {
                return {
                    innerWidth: window.innerWidth,
                    innerHeight: window.innerHeight,
                };
            });
            this.#logger.debug(`Viewport size is: ${innerWidth}x${innerHeight}`);
        }
        try {
            if (!captureElement) {
                this.#logger.debug('Capturing viewport screenshot');
                screenshot = await this.#browser.takeScreenshot();
                this.#logger.debug('Viewport screenshot is captured');
            }
            else {
                this.#logger.debug(`Checking is element ${chalk_1.default.cyan(captureElement)} fit into viewport`);
                const rects = await this.#browser.executeScript(function (selector) {
                    window.scrollTo(0, 0); // TODO Maybe we should remove same code from `resetMousePosition`
                    // eslint-disable-next-line no-var
                    var element = document.querySelector(selector);
                    if (!element)
                        return;
                    // eslint-disable-next-line no-var
                    var elementRect = element.getBoundingClientRect();
                    return {
                        elementRect: {
                            top: elementRect.top,
                            left: elementRect.left,
                            width: elementRect.width,
                            height: elementRect.height,
                        },
                        // NOTE page_Offset is used only for IE9-11
                        windowRect: {
                            // eslint-disable-next-line @typescript-eslint/no-deprecated
                            top: Math.round(window.scrollY || window.pageYOffset),
                            // eslint-disable-next-line @typescript-eslint/no-deprecated
                            left: Math.round(window.scrollX || window.pageXOffset),
                            width: window.innerWidth,
                            height: window.innerHeight,
                        },
                    };
                }, captureElement);
                const { elementRect, windowRect } = rects ?? {};
                if (!elementRect || !windowRect)
                    throw new Error(`Couldn't find element with selector: '${captureElement}'`);
                const isFitIntoViewport = elementRect.width + elementRect.left <= windowRect.width &&
                    elementRect.height + elementRect.top <= windowRect.height;
                if (isFitIntoViewport) {
                    this.#logger.debug(`Capturing ${chalk_1.default.cyan(captureElement)} with size: ${elementRect.width}x${elementRect.height}`);
                }
                else
                    this.#logger.debug(`Capturing composite screenshot image of ${chalk_1.default.cyan(captureElement)} with size: ${elementRect.width}x${elementRect.height}`);
                // const element = await browser.findElement(By.css(captureElement));
                // screenshot = isFitIntoViewport
                //   ? context
                //     ? await context.captureElementScreenshot(await element.getId())
                //     : await browser.findElement(By.css(captureElement)).takeScreenshot()
                //   : // TODO pointer-events: none, need to research
                //     await takeCompositeScreenshot(browser, windowRect, elementRect);
                screenshot = isFitIntoViewport
                    ? await this.#browser.findElement(selenium_webdriver_1.By.css(captureElement)).takeScreenshot()
                    : // TODO pointer-events: none, need to research
                        await this.takeCompositeScreenshot(windowRect, elementRect);
                this.#logger.debug(`${chalk_1.default.cyan(captureElement)} is captured`);
            }
        }
        finally {
            await this.removeIgnoreStyles(ignoreStyles);
        }
        return typeof screenshot === 'string' ? Buffer.from(screenshot, 'base64') : screenshot;
    }
    waitForComplete(callback) {
        void this.#browser
            .executeAsyncScript(function (callback) {
            void window.__CREEVEY_HAS_PLAY_COMPLETED_YET__().then(callback);
        })
            .then(callback);
    }
    async selectStory(id, waitForReady = false) {
        // NOTE: Global variables might be reset after hot reload. I think it's workaround, maybe we need better solution
        await this.updateBrowserGlobalVariables();
        await this.resetMousePosition();
        this.#logger.debug(`Triggering 'SetCurrentStory' event with storyId ${chalk_1.default.magenta(id)}`);
        const result = await this.#browser.executeAsyncScript(function (storyId, shouldWaitForReady, callback) {
            if (typeof window.__CREEVEY_SELECT_STORY__ == 'undefined') {
                callback([
                    "Creevey can't switch story. This may happened if forget to add `creevey` addon to your storybook config, or storybook not loaded in browser due syntax error.",
                ]);
                return;
            }
            void window.__CREEVEY_SELECT_STORY__(storyId, shouldWaitForReady).then(callback);
        }, id, waitForReady);
        const [errorMessage, isCaptureCalled = false] = result ?? [];
        if (errorMessage)
            throw new Error(errorMessage);
        return isCaptureCalled;
    }
    async updateStoryArgs(story, updatedArgs) {
        await this.#browser.executeAsyncScript(function (storyId, updatedArgs, UPDATE_STORY_ARGS, STORY_RENDERED, callback) {
            window.__STORYBOOK_ADDONS_CHANNEL__.once(STORY_RENDERED, callback);
            window.__STORYBOOK_ADDONS_CHANNEL__.emit(UPDATE_STORY_ARGS, {
                storyId,
                updatedArgs,
            });
        }, story.id, updatedArgs, types_js_1.StorybookEvents.UPDATE_STORY_ARGS, types_js_1.StorybookEvents.STORY_RENDERED);
    }
    async loadStoriesFromBrowser() {
        const stories = await this.#browser.executeAsyncScript(function (callback) {
            void window.__CREEVEY_GET_STORIES__().then(callback);
        });
        if (!stories)
            throw new Error("Can't get stories, it seems creevey or storybook API isn't available");
        return stories;
    }
    async afterTest(test) {
        if (this.#logger.getLevel() === loglevel_1.default.levels.TRACE) {
            const output = [];
            const types = await this.#browser.manage().logs().getAvailableLogTypes();
            for (const type of types) {
                const logs = await this.#browser.manage().logs().get(type);
                output.push(logs.map((log) => JSON.stringify(log.toJSON(), null, 2)).join('\n'));
            }
            this.#logger.debug('----------', (0, utils_js_1.getTestPath)(test).join('/'), '----------\n', output.join('\n'), '\n----------------------------------------------------------------------------------------------------');
        }
    }
    static async getBrowser(browserName, gridUrl, config, options) {
        const browserConfig = config.browsers[browserName];
        const { storybookUrl: address = config.storybookUrl, limit, viewport, _storybookGlobals } = browserConfig;
        void limit;
        const browser = await buildWebdriver(browserName, gridUrl, config, options);
        if (!browser)
            return null;
        const internalBrowser = new InternalBrowser(browser, options.port, logger_js_1.logger);
        try {
            if (utils_js_1.isShuttingDown.current)
                return null;
            const done = await internalBrowser.init({
                browserName,
                gridUrl,
                viewport,
                storybookUrl: address,
                storybookGlobals: _storybookGlobals,
                resolveStorybookUrl: config.resolveStorybookUrl,
            });
            return done ? internalBrowser : null;
        }
        catch (originalError) {
            void internalBrowser.closeBrowser();
            const message = originalError instanceof Error ? originalError.message : originalError;
            const error = new Error(`Can't load storybook root page by URL ${await browser.getCurrentUrl()}: ${message}`);
            if (originalError instanceof Error)
                error.stack = originalError.stack;
            logger_js_1.logger.error(error);
            return null;
        }
    }
    async init({ browserName, gridUrl, viewport, storybookUrl, storybookGlobals, resolveStorybookUrl, }) {
        const sessionId = (await this.#browser.getSession()).getId();
        let browserHost = '';
        try {
            const { Name } = await getSessionData(gridUrl, sessionId);
            if (typeof Name == 'string')
                browserHost = Name;
        }
        catch {
            /* noop */
        }
        this.#logger = loglevel_1.default.getLogger(sessionId);
        loglevel_plugin_prefix_1.default.apply(this.#logger, {
            format(level) {
                const levelColor = logger_js_1.colors[level.toUpperCase()];
                return `[${browserName}:${chalk_1.default.gray(sessionId)}] ${levelColor(level)} =>`;
            },
        });
        this.#logger.debug(`Connected successful with ${chalk_1.default.green(browserHost)}`);
        return await (0, utils_js_1.runSequence)([
            () => this.#browser.manage().setTimeouts({ pageLoad: 10000, script: 60000 }),
            () => this.openStorybookPage(storybookUrl, resolveStorybookUrl),
            () => this.waitForStorybook(),
            () => this.updateStorybookGlobals(storybookGlobals),
            () => this.resolveCreeveyHost(),
            () => this.updateBrowserGlobalVariables(),
            // NOTE: Selenium draws automation toolbar with some delay after webdriver initialization
            // NOTE: So if we resize window right after getting webdriver instance we might get situation
            // NOTE: When the toolbar appears after resize and final viewport size become smaller than we set
            () => this.resizeViewport(viewport),
            () => {
                this.keepAlive();
            },
        ], () => !this.#isShuttingDown);
    }
    async openStorybookPage(storybookUrl, resolver) {
        if (!webdriver_js_1.LOCALHOST_REGEXP.test(storybookUrl)) {
            return this.#browser.get((0, webdriver_js_1.appendIframePath)(storybookUrl));
        }
        try {
            if (resolver) {
                this.#logger.debug('Resolving storybook url with custom resolver');
                const resolvedUrl = await resolver();
                this.#logger.debug(`Resolver storybook url ${resolvedUrl}`);
                await this.#browser.get((0, webdriver_js_1.appendIframePath)(resolvedUrl));
            }
            else {
                // NOTE: getUrlChecker already calls `browser.get` so we don't need another one
                await (0, webdriver_js_1.resolveStorybookUrl)((0, webdriver_js_1.appendIframePath)(storybookUrl), (url) => this.checkUrl(url), this.#logger);
            }
        }
        catch (error) {
            this.#logger.error('Failed to resolve storybook URL', error instanceof Error ? error.message : '');
            throw error;
        }
    }
    async checkUrl(url) {
        try {
            // NOTE: Before trying a new url, reset the current one
            this.#logger.debug(`Opening ${chalk_1.default.magenta('about:blank')} page`);
            await openUrlAndWaitForPageSource(this.#browser, 'about:blank', (source) => !source.includes('<body></body>'));
            this.#logger.debug(`Opening ${chalk_1.default.magenta(url)} and checking the page source`);
            const source = await openUrlAndWaitForPageSource(this.#browser, url, 
            // NOTE: IE11 can return only `head` without body
            (source) => source.length == 0 || !/<body([^>]*>).+<\/body>/s.test(source));
            // NOTE: This is the most optimal way to check if we in storybook or not
            // We don't use any page load strategies except `NONE`
            // because other add significant delay and some of them don't work in earlier chrome versions
            // Browsers always load page successful even it's failed
            // So we just check `root` element
            this.#logger.debug(`Checking ${chalk_1.default.cyan(`#${webdriver_js_1.storybookRootID}`)} existence on ${chalk_1.default.magenta(url)}`);
            return source.includes(`id="${webdriver_js_1.storybookRootID}"`);
        }
        catch {
            return false;
        }
    }
    async waitForStorybook() {
        this.#logger.debug('Waiting for `setStories` event to make sure that storybook is initiated');
        const isTimeout = await Promise.race([
            new Promise((resolve) => {
                setTimeout(() => {
                    resolve(true);
                }, 60000);
            }),
            (async () => {
                let wait = true;
                do {
                    try {
                        // TODO Research a different way to ensure storybook is initiated
                        wait = await this.#browser.executeScript(function (SET_GLOBALS) {
                            // TODO Maybe use
                            // import { global } from '@storybook/global';
                            // global.IS_STORYBOOK
                            if (typeof window.__STORYBOOK_ADDONS_CHANNEL__ == 'undefined')
                                return true;
                            if (window.__STORYBOOK_ADDONS_CHANNEL__.last(SET_GLOBALS) == undefined)
                                return true;
                            return false;
                        }, types_js_1.StorybookEvents.SET_GLOBALS);
                    }
                    catch (e) {
                        this.#logger.debug('An error has been caught during the script:', e);
                    }
                } while (wait);
                return false;
            })(),
        ]);
        // TODO Change the message to describe a reason why it might happen
        if (isTimeout)
            throw new Error('Failed to wait `setStories` event');
    }
    async updateStorybookGlobals(globals) {
        if (!globals)
            return;
        this.#logger.debug('Applying storybook globals');
        await this.#browser.executeScript(function (globals) {
            window.__CREEVEY_UPDATE_GLOBALS__(globals);
        }, globals);
    }
    async resolveCreeveyHost() {
        const addresses = (0, webdriver_js_1.getAddresses)();
        this.#serverHost = await this.#browser.executeAsyncScript(function (hosts, port, callback) {
            void Promise.all(hosts.map(function (host) {
                return Promise.race([
                    // eslint-disable-next-line @typescript-eslint/restrict-plus-operands
                    fetch('http://' + host + ':' + port + '/ping').then(function (response) {
                        return response.text();
                    }),
                    new Promise((_resolve, reject) => {
                        setTimeout(reject, 5000);
                    }),
                ])
                    .then(function (pong) {
                    return pong == 'pong' ? host : null;
                })
                    .catch(function () {
                    return null;
                });
            })).then(function (hosts) {
                callback(hosts.find(function (host) {
                    return host != null;
                }));
            });
        }, addresses, this.#serverPort);
        if (this.#serverHost == null)
            throw new Error("Can't reach creevey server from a browser");
    }
    async updateBrowserGlobalVariables() {
        await this.#browser.executeScript(function (workerId, creeveyHost, creeveyPort) {
            window.__CREEVEY_WORKER_ID__ = workerId;
            window.__CREEVEY_SERVER_HOST__ = creeveyHost;
            window.__CREEVEY_SERVER_PORT__ = creeveyPort;
        }, process.pid, this.#serverHost, this.#serverPort);
    }
    async resizeViewport(viewport) {
        if (!viewport)
            return;
        const windowRect = await this.#browser.manage().window().getRect();
        const { innerWidth, innerHeight } = await this.#browser.executeScript(function () {
            return {
                innerWidth: window.innerWidth,
                innerHeight: window.innerHeight,
            };
        });
        this.#logger.debug(`Resizing viewport from ${innerWidth}x${innerHeight} to ${viewport.width}x${viewport.height}`);
        const dWidth = windowRect.width - innerWidth;
        const dHeight = windowRect.height - innerHeight;
        await this.#browser
            .manage()
            .window()
            .setRect({
            width: viewport.width + dWidth,
            height: viewport.height + dHeight,
        });
    }
    async resetMousePosition() {
        this.#logger.debug('Resetting mouse position to the top-left corner');
        const browserName = (await this.#browser.getCapabilities()).getBrowserName();
        const [browserVersion] = (await this.#browser.getCapabilities()).getBrowserVersion()?.split('.') ??
            (await this.#browser.getCapabilities()).get('version')?.split('.') ??
            [];
        // NOTE Reset mouse position to support keweb selenium grid browser versions
        if (browserName == 'chrome' && browserVersion == '70') {
            const { top, left, width, height } = await this.#browser.executeScript(function () {
                const bodyRect = document.body.getBoundingClientRect();
                return {
                    top: bodyRect.top,
                    left: bodyRect.left,
                    width: bodyRect.width,
                    height: bodyRect.height,
                };
            });
            // NOTE Bridge mode doesn't support `Origin.VIEWPORT`, move mouse relative
            await this.#browser
                .actions({ bridge: true })
                .move({
                origin: this.#browser.findElement(selenium_webdriver_1.By.css('body')),
                x: Math.ceil((-1 * width) / 2) - left,
                y: Math.ceil((-1 * height) / 2) - top,
            })
                .perform();
        }
        else if (browserName == 'firefox' && browserVersion == '61') {
            // NOTE Firefox for some reason moving by 0 x 0 move cursor in bottom left corner :sad:
            await this.#browser.actions().move({ origin: selenium_webdriver_1.Origin.VIEWPORT, x: 0, y: 1 }).perform();
        }
        else {
            // NOTE IE don't emit move events until force window focus or connect by RDP on virtual machine
            await this.#browser.actions().move({ origin: selenium_webdriver_1.Origin.VIEWPORT, x: 0, y: 0 }).perform();
        }
    }
    async insertIgnoreStyles(ignoreElements) {
        const ignoreSelectors = Array.prototype.concat(ignoreElements).filter(Boolean);
        if (!ignoreSelectors.length)
            return null;
        this.#logger.debug('Hiding ignored elements before capturing');
        return await this.#browser.executeScript(function (ignoreSelectors) {
            return window.__CREEVEY_INSERT_IGNORE_STYLES__(ignoreSelectors);
        }, ignoreSelectors);
    }
    async takeCompositeScreenshot(windowRect, elementRect) {
        const screens = [];
        const isScreenshotWithoutScrollBar = !(await this.hasScrollBar());
        const scrollBarWidth = await this.getScrollBarWidth();
        // NOTE Sometimes viewport has been scrolled somewhere
        const normalizedElementRect = {
            left: elementRect.left - windowRect.left,
            right: elementRect.left + elementRect.width - windowRect.left,
            top: elementRect.top - windowRect.top,
            bottom: elementRect.top + elementRect.height - windowRect.top,
        };
        const isFitHorizontally = windowRect.width >= elementRect.width + normalizedElementRect.left;
        const isFitVertically = windowRect.height >= elementRect.height + normalizedElementRect.top;
        const viewportWidth = windowRect.width - (isFitVertically ? 0 : scrollBarWidth);
        const viewportHeight = windowRect.height - (isFitHorizontally ? 0 : scrollBarWidth);
        const cols = Math.ceil(elementRect.width / viewportWidth);
        const rows = Math.ceil(elementRect.height / viewportHeight);
        const xOffset = Math.round(isFitHorizontally ? normalizedElementRect.left : Math.max(0, cols * viewportWidth - elementRect.width));
        const yOffset = Math.round(isFitVertically ? normalizedElementRect.top : Math.max(0, rows * viewportHeight - elementRect.height));
        for (let row = 0; row < rows; row += 1) {
            for (let col = 0; col < cols; col += 1) {
                const dx = Math.min(viewportWidth * col + normalizedElementRect.left, Math.max(0, normalizedElementRect.right - viewportWidth));
                const dy = Math.min(viewportHeight * row + normalizedElementRect.top, Math.max(0, normalizedElementRect.bottom - viewportHeight));
                await this.#browser.executeScript(function (x, y) {
                    window.scrollTo(x, y);
                }, dx, dy);
                screens.push(await this.#browser.takeScreenshot());
            }
        }
        const images = screens.map((s) => Buffer.from(s, 'base64')).map((b) => pngjs_1.PNG.sync.read(b));
        const compositeImage = new pngjs_1.PNG({ width: Math.round(elementRect.width), height: Math.round(elementRect.height) });
        for (let y = 0; y < compositeImage.height; y += 1) {
            for (let x = 0; x < compositeImage.width; x += 1) {
                const col = Math.floor(x / viewportWidth);
                const row = Math.floor(y / viewportHeight);
                const isLastCol = cols - col == 1;
                const isLastRow = rows - row == 1;
                const scrollOffset = isFitVertically || isScreenshotWithoutScrollBar ? 0 : scrollBarWidth;
                const i = (y * compositeImage.width + x) * 4;
                const j = 
                // NOTE compositeImage(x, y) => image(x, y)
                ((y % viewportHeight) * (viewportWidth + scrollOffset) + (x % viewportWidth)) * 4 +
                    // NOTE Offset for last row/col image
                    (isLastRow ? yOffset * (viewportWidth + scrollOffset) * 4 : 0) +
                    (isLastCol ? xOffset * 4 : 0);
                const image = images[row * cols + col];
                compositeImage.data[i + 0] = image.data[j + 0];
                compositeImage.data[i + 1] = image.data[j + 1];
                compositeImage.data[i + 2] = image.data[j + 2];
                compositeImage.data[i + 3] = image.data[j + 3];
            }
        }
        return pngjs_1.PNG.sync.write(compositeImage);
    }
    async removeIgnoreStyles(ignoreStyles) {
        if (ignoreStyles) {
            this.#logger.debug('Revert hiding ignored elements');
            await this.#browser.executeScript(function (ignoreStyles) {
                window.__CREEVEY_REMOVE_IGNORE_STYLES__(ignoreStyles);
            }, ignoreStyles);
        }
    }
    // NOTE Firefox and Safari take viewport screenshot without scrollbars
    async hasScrollBar() {
        const browserName = (await this.#browser.getCapabilities()).getBrowserName();
        const [browserVersion] = (await this.#browser.getCapabilities()).getBrowserVersion()?.split('.') ?? [];
        return (browserName != 'Safari' &&
            // NOTE This need to work with keweb selenium grid
            !(browserName == 'firefox' && browserVersion == '61'));
    }
    async getScrollBarWidth() {
        const scrollBarWidth = await this.#browser.executeScript(function () {
            // eslint-disable-next-line no-var
            var div = document.createElement('div');
            div.innerHTML = 'a'; // NOTE: In IE clientWidth is 0 if this div is empty.
            div.style.overflowY = 'scroll';
            document.body.appendChild(div);
            // eslint-disable-next-line no-var
            var widthDiff = div.offsetWidth - div.clientWidth;
            document.body.removeChild(div);
            return widthDiff;
        });
        return scrollBarWidth;
    }
    keepAlive() {
        this.#keepAliveInterval = setInterval(() => {
            // NOTE Simple way to keep session alive
            void this.#browser.getCurrentUrl().then((url) => {
                logger_js_1.logger.debug('current url', chalk_1.default.magenta(url));
            });
        }, 10 * 1000);
    }
}
exports.InternalBrowser = InternalBrowser;
//# sourceMappingURL=internal.js.map