"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CreeveyWebdriverBase = exports.LOCALHOST_REGEXP = exports.storybookRootID = void 0;
exports.resolveStorybookUrl = resolveStorybookUrl;
exports.appendIframePath = appendIframePath;
exports.getAddresses = getAddresses;
const chalk_1 = __importDefault(require("chalk"));
const os_1 = require("os");
const logger_js_1 = require("./logger.js");
const types_js_1 = require("../types.js");
const messages_js_1 = require("./messages.js");
exports.storybookRootID = 'storybook-root';
exports.LOCALHOST_REGEXP = /(localhost|127\.0\.0\.1)/i;
const DOCKER_INTERNAL = 'host.docker.internal';
async function resolveStorybookUrl(storybookUrl, checkUrl, logger = logger_js_1.logger) {
    logger.debug('Resolving storybook url');
    const addresses = getAddresses();
    for (const ip of addresses) {
        const resolvedUrl = storybookUrl.replace(exports.LOCALHOST_REGEXP, ip);
        logger.debug(`Checking storybook availability on ${chalk_1.default.magenta(resolvedUrl)}`);
        if (await checkUrl(resolvedUrl)) {
            logger.debug(`Resolved storybook url ${chalk_1.default.magenta(resolvedUrl)}`);
            return resolvedUrl;
        }
    }
    const error = new Error('Please specify `storybookUrl` with IP address that accessible from remote browser');
    error.name = 'ResolveUrlError';
    throw error;
}
function appendIframePath(url) {
    return `${url.replace(/\/$/, '')}/iframe.html`;
}
function getAddresses() {
    // TODO Check if docker is used
    return [DOCKER_INTERNAL].concat(...Object.values((0, os_1.networkInterfaces)())
        .filter(types_js_1.isDefined)
        .map((network) => network.filter((info) => info.family == 'IPv4').map((info) => info.address)));
}
class CreeveyWebdriverBase {
    async switchStory(story, context, logger) {
        const { id, title, name, parameters } = story;
        const { captureElement = `#${exports.storybookRootID}`, waitForReady, ignoreElements, } = (parameters.creevey ?? {});
        logger.debug(`Switching to story ${chalk_1.default.cyan(title)}/${chalk_1.default.cyan(name)} by id ${chalk_1.default.magenta(id)}`);
        let storyPlayResolver;
        let waitForComplete = new Promise((resolve) => (storyPlayResolver = resolve));
        const unsubscribe = (0, messages_js_1.subscribeOn)('stories', (message) => {
            if (message.type != 'capture')
                return;
            const { payload = {}, payload: { imageName } = {} } = message;
            void this.takeScreenshot(payload.captureElement ?? captureElement, payload.ignoreElements ?? ignoreElements).then((screenshot) => {
                context.screenshots.push({ imageName, screenshot });
                this.waitForComplete(storyPlayResolver);
                (0, messages_js_1.emitStoriesMessage)({ type: 'capture' });
            });
        });
        const isCaptureCalled = await this.selectStory(id, waitForReady);
        if (isCaptureCalled) {
            logger.debug(`Capturing screenshots from ${chalk_1.default.magenta(id)} story's \`play()\` function`);
            while (!(await waitForComplete)) {
                waitForComplete = new Promise((resolve) => (storyPlayResolver = resolve));
            }
        }
        unsubscribe();
        if (isCaptureCalled)
            logger.debug(`Story ${chalk_1.default.magenta(id)} completed capturing`);
        else
            logger.debug(`Story ${chalk_1.default.magenta(id)} ready for capturing`);
        return Object.assign({
            takeScreenshot: () => this.takeScreenshot(captureElement, ignoreElements),
            updateStoryArgs: (updatedArgs) => this.updateStoryArgs(story, updatedArgs),
            captureElement,
        }, context);
    }
}
exports.CreeveyWebdriverBase = CreeveyWebdriverBase;
//# sourceMappingURL=webdriver.js.map