"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.start = start;
const chai_1 = __importDefault(require("chai"));
const chalk_1 = __importDefault(require("chalk"));
const loglevel_1 = __importDefault(require("loglevel"));
const events_1 = __importDefault(require("events"));
const types_js_1 = require("../../types.js");
const messages_js_1 = require("../messages.js");
const chai_image_js_1 = __importDefault(require("./chai-image.js"));
const match_image_js_1 = require("./match-image.js");
const stories_js_1 = require("../stories.js");
const logger_js_1 = require("../logger.js");
const utils_js_1 = require("../utils.js");
async function getTestsFromStories(config, browserName, webdriver) {
    const testsById = new Map();
    const tests = await (0, stories_js_1.loadTestsFromStories)([browserName], (listener) => config.storiesProvider(config, listener, webdriver), (testsDiff) => {
        Object.entries(testsDiff).forEach(([id, newTest]) => {
            if (newTest)
                testsById.set(id, newTest);
            else
                testsById.delete(id);
        });
    });
    Object.values(tests)
        .filter(types_js_1.isDefined)
        .forEach((test) => testsById.set(test.id, test));
    return testsById;
}
function runHandler(browserName, images, error) {
    // TODO How handle browser corruption?
    if ((0, types_js_1.isImageError)(error)) {
        if (typeof error.images == 'string') {
            const image = images[browserName];
            if (image)
                image.error = error.images;
        }
        else {
            const imageErrors = error.images ?? {};
            Object.keys(imageErrors).forEach((imageName) => {
                const image = images[imageName];
                if (image)
                    image.error = imageErrors[imageName];
            });
        }
    }
    if (error || Object.values(images).some((image) => image?.error != null)) {
        const errorMessage = serializeError(error);
        const isUnexpectedError = hasTimeout(errorMessage) ||
            hasDisconnected(errorMessage) ||
            Object.values(images).some((image) => hasTimeout(image?.error));
        if (isUnexpectedError)
            (0, messages_js_1.emitWorkerMessage)({ type: 'error', payload: { subtype: 'unknown', error: errorMessage } });
        else
            (0, messages_js_1.emitTestMessage)({
                type: 'end',
                payload: {
                    status: 'failed',
                    images,
                    error: errorMessage,
                },
            });
    }
    else {
        (0, messages_js_1.emitTestMessage)({ type: 'end', payload: { status: 'success', images } });
    }
}
async function setupWebdriver(webdriver) {
    if ((await webdriver.openBrowser(true)) == null) {
        logger_js_1.logger.error('Failed to start browser');
        (0, messages_js_1.emitWorkerMessage)({
            type: 'error',
            payload: { subtype: 'browser', error: 'Failed to start browser' },
        });
        return;
    }
    const sessionId = await webdriver.getSessionId();
    return [sessionId, webdriver];
}
function serializeError(error) {
    if (!error)
        return 'Unknown error';
    if (error instanceof Error)
        return error.stack ?? error.message;
    return typeof error === 'object' ? JSON.stringify(error) : error;
}
function hasDisconnected(str) {
    return str?.toLowerCase().includes('disconnected') ?? false;
}
function hasTimeout(str) {
    return str?.toLowerCase().includes('timeout') ?? false;
}
async function start(browser, gridUrl, config, options) {
    let retries = 0;
    const imagesContext = {
        attachments: [],
        testFullPath: [],
        images: {},
    };
    const Webdriver = config.webdriver;
    const [sessionId, webdriver] = (await setupWebdriver(new Webdriver(browser, gridUrl, config, options))) ?? [];
    if (!webdriver || !sessionId)
        return;
    const workerLogger = loglevel_1.default.getLogger(`${browser}:${chalk_1.default.gray(sessionId)}`);
    const reporterOptions = {
        ...config.reporterOptions,
        creevey: {
            sessionId,
            reportDir: config.reportDir,
            browserName: browser,
            get willRetry() {
                return retries < config.maxRetries;
            },
            get images() {
                return imagesContext.images;
            },
        },
    };
    class FakeRunner extends events_1.default {
    }
    const runner = new FakeRunner();
    const Reporter = config.reporter;
    new Reporter(runner, { reporterOptions });
    const { matchImage, matchImages } = options.odiff
        ? (0, match_image_js_1.getOdiffMatchers)(imagesContext, config)
        : await (0, match_image_js_1.getMatchers)(imagesContext, config);
    chai_1.default.use((0, chai_image_js_1.default)(matchImage, matchImages, workerLogger));
    const tests = await (async () => {
        try {
            return await getTestsFromStories(config, browser, webdriver);
        }
        catch (error) {
            workerLogger.error('Failed to get tests from stories:', error);
            (0, messages_js_1.emitWorkerMessage)({
                type: 'error',
                payload: { subtype: 'browser', error: serializeError(error) },
            });
            return null;
        }
    })();
    if (!tests)
        return;
    (0, messages_js_1.subscribeOn)('test', (message) => {
        if (message.type != 'start')
            return;
        const test = tests.get(message.payload.id);
        if (!test) {
            const error = `Test with id ${message.payload.id} not found`;
            workerLogger.error(error);
            (0, messages_js_1.emitWorkerMessage)({
                type: 'error',
                payload: { subtype: 'test', error },
            });
            return;
        }
        const baseContext = {
            browserName: browser,
            // @ts-expect-error We defined separate d.ts declarations for each webdriver
            // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
            webdriver: webdriver.browser,
            screenshots: [],
            matchImage: matchImage,
            matchImages: matchImages,
            // NOTE: Deprecated
            expect: chai_1.default.expect,
        };
        imagesContext.attachments = [];
        imagesContext.testFullPath = (0, utils_js_1.getTestPath)(test);
        imagesContext.images = {};
        retries = message.payload.retries;
        let error = undefined;
        const fakeSuite = {
            title: test.storyPath.slice(0, -1).join('/'),
            fullTitle: () => fakeSuite.title,
            titlePath: () => [fakeSuite.title],
            tests: [],
        };
        const fakeTest = {
            parent: fakeSuite,
            title: [test.story.name, test.testName, test.browser].filter(types_js_1.isDefined).join('/'),
            fullTitle: () => (0, utils_js_1.getTestPath)(test).join('/'),
            titlePath: () => (0, utils_js_1.getTestPath)(test),
            currentRetry: () => retries,
            retires: () => config.maxRetries,
            slow: () => 1000,
        };
        fakeSuite.tests.push(fakeTest);
        void (async () => {
            runner.emit(types_js_1.TEST_EVENTS.RUN_BEGIN);
            runner.emit(types_js_1.TEST_EVENTS.TEST_BEGIN, fakeTest);
            const start = Date.now();
            try {
                await Promise.race([
                    new Promise((reject) => setTimeout(() => {
                        reject(`Timeout of ${config.testTimeout}ms exceeded`);
                    }, config.testTimeout)),
                    (async () => {
                        const context = await webdriver.switchStory(test.story, baseContext, workerLogger);
                        await test.fn(context);
                    })(),
                ]);
            }
            catch (testError) {
                error = testError;
                fakeTest.err = error;
            }
            const duration = Date.now() - start;
            fakeTest.attachments = imagesContext.attachments;
            fakeTest.state = error ? 'failed' : 'passed';
            fakeTest.duration = duration;
            fakeTest.speed = duration > fakeTest.slow() ? 'slow' : duration / 2 > fakeTest.slow() ? 'medium' : 'fast';
            if (error) {
                runner.emit(types_js_1.TEST_EVENTS.TEST_FAIL, fakeTest, error);
            }
            else {
                runner.emit(types_js_1.TEST_EVENTS.TEST_PASS, fakeTest);
            }
            runner.emit(types_js_1.TEST_EVENTS.TEST_END, fakeTest);
            runner.emit(types_js_1.TEST_EVENTS.RUN_END);
            await webdriver.afterTest(test);
            runHandler(baseContext.browserName, imagesContext.images, error);
        })().catch((error) => {
            workerLogger.error('Unexpected error:', error);
            (0, messages_js_1.emitWorkerMessage)({
                type: 'error',
                payload: { subtype: 'test', error: serializeError(error) },
            });
        });
    });
    workerLogger.info('Browser is ready');
    (0, messages_js_1.emitWorkerMessage)({ type: 'ready' });
}
//# sourceMappingURL=start.js.map